<?php
/**
 * Plugin Name: Easy Booking Hour Extended
 * Description: WooCommerce Easy Booking add-on for hour bookings, stocks and availability
 * Version: 1.2.0
 * Text domain: j_easy_booking_extended
 * Licence : GPLv2 or later
 *
 * @Author: José Elías Gutierrez
 * @Date  :   2017-04-10 10:51:30
 * @Last  Modified by:   José Elías Gutierrez
 * @Last  Modified time: 2017-06-19 14:49:42
 */

if (!defined('ABSPATH')) {
    exit; // Exit if accessed directly
}

if (!class_exists('J_Easy_Booking_Extended')) :


    define('JEBE_PAGE_ENVIO_ALQUILER_ID', 19591);
    define('JEBE_PAGE_GARANTIA_ALQUILER_ID', 19593);
    define('JEBE_PAGE_SEGURO_ALQUILER_ID', 19589);

    class J_Easy_Booking_Extended
    {

        protected static $_instance = null;

        public static function instance()
        {

            if (is_null(self::$_instance)) {
                self::$_instance = new self();
            }

            return self::$_instance;
        }

        public function __construct()
        {

            // Check if WooCommerce Easy Booking is active
            if ($this->jebe_easy_booking_is_active()) {

                $plugin = plugin_basename(__FILE__);

                // Init plugin
                add_action('plugins_loaded', [$this, 'jebe_init'], 20);

                # TODO: move actions to proper function to avoid unused functions querying

                // Add custom fields to product variations (price per hour, price per day, price per weekend, holiday price)
                add_action('easy_booking_after_variation_booking_options', 'jebe_add_variation_price_fields', 10, 2);


                //  Save variations price fields
                add_action('woocommerce_save_product_variation', 'jebe_booking_price_fields_save', 10, 2);

                add_action('woocommerce_before_add_to_cart_button', [$this, 'remove_actions'], 1);
                add_action('woocommerce_before_add_to_cart_button', 'jebe_display_for_variable_product', 20);
                // css file
                add_action('woocommerce_before_add_to_cart_button', 'jebe_include_styles', 20);

                // tab filters
                add_filter('woocommerce_product_tabs', [$this, 'jebe_replace_tabs'], 90, 1);


                // Product page actions

                # wceb returns html. We need to override it.
                $this->remove_ajax_actions();

                add_action('wp_ajax_add_new_price', [$this, 'jebe_get_new_price']);
                add_action('wp_ajax_nopriv_add_new_price', [$this, 'jebe_get_new_price']);

                # Cart actions
                add_filter('woocommerce_add_to_cart_validation', 'jebe_check_hours_before_add_to_cart', 30, 5);
                add_filter('woocommerce_add_cart_item_data', 'jebe_add_cart_item_data', 30, 3);
                add_filter('woocommerce_get_cart_item_from_session', 'jebe_get_cart_item_from_session', 99, 2);
                add_filter('woocommerce_get_item_data', 'jebe_get_item_data', 20, 2);
                add_filter('woocommerce_get_price_html', [$this, 'jebe_add_price_html'], 90, 2);

                # checkout actions
                if (version_compare(WC_VERSION, '2.7', '<')) {
                    add_action('woocommerce_add_order_item_meta', 'jebe_add_order_meta', 10, 2);
                    add_filter('woocommerce_order_items_meta_display', 'jebe_display_booking_item_meta', 10, 2);
                } else {
                    add_action('woocommerce_checkout_create_order_line_item', 'jebe_order_line_item', 20, 3);
                    add_filter('woocommerce_display_item_meta', 'jebe_display_item_meta', 20, 3);
                }
                add_filter('woocommerce_hidden_order_itemmeta', 'jebe_hide_formatted_date', 20, 1);


                if (is_admin()) {
                    # display hours in order page
                    add_action('woocommerce_before_order_itemmeta', 'jebe_order_display_product_dates', 20, 3);

                }

                add_filter('woocommerce_get_availability', [$this, 'jebe_remove_availability_messages'], 90, 2);
            }
        }

        public function jebe_replace_tabs($tabs)
        {

            global $product;

            if (wceb_is_bookable($product)) {

                $pageId = icl_object_id(JEBE_PAGE_GARANTIA_ALQUILER_ID, 'page');

                $tabs['garantia']['title'] = get_the_title($pageId);
                $tabs['garantia']['callback'] = function () {
                    $pageId = icl_object_id(JEBE_PAGE_GARANTIA_ALQUILER_ID, 'page');
                    if ($pageId) {
                        echo apply_filters('the_content', get_post($pageId)->post_content);
                    }
                };

                $pageId = icl_object_id(JEBE_PAGE_ENVIO_ALQUILER_ID, 'page');
                $tabs['envio_entrega']['title'] = get_the_title($pageId);
                $tabs['envio_entrega']['callback'] = function () {
                    $pageId = icl_object_id(JEBE_PAGE_ENVIO_ALQUILER_ID, 'page');
                    if ($pageId) {
                        echo apply_filters('the_content', get_post($pageId)->post_content);
                    }
                };

                $pageId = icl_object_id(JEBE_PAGE_SEGURO_ALQUILER_ID, 'page');
                $tabs['seguro'] = [
                    'title'    => get_the_title($pageId),
                    'priority' => 50,
                    'callback' => function () {
                        $pageId = icl_object_id(JEBE_PAGE_SEGURO_ALQUILER_ID, 'page');
                        if ($pageId) {
                            echo apply_filters('the_content', get_post($pageId)->post_content);
                        }
                    },
                ];

            }

            return $tabs;
        }

        public function jebe_add_price_html($content, $product)
        {

            if (!$product)
                return;

            if (wceb_is_bookable($product)) {

                $content = __('Desde ', 'j_easy_booking_extended') . $content . '<span class="jebe-min-booking-range tax-includes"> ' . __('* Mínimo 2 horas', 'j_easy_booking_extended') . '</span>';
            }

            return $content;

        }

        public function jebe_remove_availability_messages($availability_html, $product)
        {

            if (wceb_is_bookable($product))
                $availability_html = '';

            return $availability_html;
        }

        public function get_max_bookable_qty($variation_id, $data)
        {

            $bookable = true;

            $variation_obj = new WC_Product_variation($variation_id);
            $stock = $variation_obj->get_stock_quantity();

            $start_date = new DateTime($data['start'] . ' ' . $data['start_hour']);
            $end_date = new DateTime($data['end'] . ' ' . $data['end_hour']);

            $bookings = $this->get_variation_bookings($variation_id);

            $overlap = [];

            // check if booking dont overlap
            foreach ($bookings['booked'] as $key => $booking) {

                $booking_start_date = new DateTime($booking['start_date'] . ' ' . $booking['start_hour']);
                $booking_end_date = new DateTime($booking['end_date'] . ' ' . $booking['end_hour']);

                if (!($end_date < $booking_start_date || $start_date > $booking_end_date)) {
                    $overlap[] = $booking;
                }

            }

            // no overlap, so all stock is bookable
            if (count($overlap) == 0)
                return $stock;

            // overlaps, let's check qtys
            // make 1 hour ranges because booking start maybe at anytime and check for everyone. If all of them has qty available, is bookable.
            $interval = new DateInterval('PT1H');
            // if start and end are the same, add one hour to end to make a valid date period
            if ($start_date == $end_date)
                $end_date->add(new DateInterval('PT1H'));
            $daterange = new DatePeriod($start_date, $interval, $end_date);

            $booked_per_hour = [];

            foreach ($daterange as $key => $date) {
                foreach ($overlap as $key => $booking) {
                    $booking_start_date = new DateTime($booking['start_date'] . ' ' . $booking['start_hour']);
                    $booking_end_date = new DateTime($booking['end_date'] . ' ' . $booking['end_hour']);

                    if ($date >= $booking_start_date && $date <= $booking_end_date) {

                        $date_key = $date->format('Y-m-d H:i');
                        if (!array_key_exists($date_key, $booked_per_hour))
                            $booked_per_hour[$date_key] = 0;

                        $booked_per_hour[$date_key] += $booking['qty'];
                    }
                }
            }

            // seek and return the minimun available qty
            $available = $stock;
            foreach ($booked_per_hour as $key => $booked) {
                $diff = $stock - $booked;

                if ($diff < $available) {
                    $available = $diff;
                }
            }

            return $available;
        }

        public function jebe_get_days_availability()
        {
            global $product;

            $availabilities = [];

            $settings = get_option('jebe_settings');
            $min_bookable_hour = $settings['jebe_min_bookable_hour'];
            $max_bookable_hour = $settings['jebe_max_bookable_hour'];

            if (wceb_is_bookable($product)) {

                $variation_ids = $product->get_children();

                if ($variation_ids) {

                    $start_date = DateTime::createFromFormat('H:i', '00:00');
                    $end_date = new DateTime('now');
                    $end_date->add(new DateInterval('P1D'));

                    foreach ($variation_ids as $variation_id) {

                        $bookings = $this->get_variation_bookings($variation_id);


                        if (count($bookings['booked']) > 0) {
                            $latest_booking = $bookings['booked'][0];

                            foreach ($bookings['booked'] as $key => $booking) {
                                if ($booking['end_date'] >= $latest_booking['end_date'] &&
                                    $booking['end_hour'] > $latest_booking['end_hour']
                                ) {
                                    $latest_booking = $booking;
                                }
                            }

                            if ($latest_booking['end_date'] >= date('Y-m-d')) {
                                $end_date = DateTime::createFromFormat('Y-m-d H:i', $latest_booking['end_date'] . ' ' . $latest_booking['end_hour']);
                                // add 1 day because we need info for later hours than the booking
                                $end_date->add(new DateInterval('P1D'));
                            }
                        }

                        $interval = new DateInterval('PT1H');
                        $date_range = new DatePeriod($start_date, $interval, $end_date);

                        foreach ($date_range as $date) {
                            if ((int)$date->format('H') >= $min_bookable_hour &&
                                (int)$date->format('H') <= $max_bookable_hour
                            ) {


                                $availabilities[$variation_id]['hour'][$date->format('Y-m-d H:i')] = $this->get_max_bookable_qty(
                                    $variation_id,
                                    [
                                        'start'      => $date->format('Y-m-d'),
                                        'start_hour' => $date->format('H:i'),
                                        'end'        => $date->format('Y-m-d'),
                                        'end_hour'   => $date->format('H:i'),
                                    ]);
                            }
                        }

                        $interval = new DateInterval('P1D');

                        $date_range = new DatePeriod($start_date, $interval, $end_date);

                        foreach ($date_range as $date) {

                            $availabilities[$variation_id]['day'][$date->format('Y-m-d')] = $this->get_max_bookable_qty(
                                $variation_id,
                                [
                                    'start'      => $date->format('Y-m-d'),
                                    'start_hour' => $date->format('H:i'),
                                    'end'        => $date->add(new DateInterval('P1D'))->format('Y-m-d'),
                                    'end_hour'   => $date->format('H:i'),
                                ]
                            );

                        }

                    }
                }
            }

            return $availabilities;
        }


        public function get_variation_bookings($variation_id)
        {
            global $wpdb;

            $options = get_option('jebe_settings');
            $attribute_name = $options['jebe_stock_attribute'];

            $variation = wc_get_product($variation_id);

            $siblings = [];

            if ($attribute_name) {
                // get variation parent
                $parent = $variation->parent->id;
                // get attribute value for variation
                $attributes = $variation->get_variation_attributes();
                $attribute_value = $attributes['attribute_pa_' . $attribute_name];
                // get all variations with parent & attribute value

                foreach ($variation->parent->get_children() as $child_id) {
                    $child = wc_get_product($child_id);
                    $child_attributes = $child->get_variation_attributes();
                    if ($attribute_value == $child_attributes['attribute_pa_' . $attribute_name]) {
                        $siblings[] = $child_id;
                    }
                }
            }


            $data = [
                'variation_id' => $variation_id,
                'booked'       => [],
                'attr'         => $attribute_name,
                'attr_value'   => $attribute_value,
                'siblings'     => $siblings,
                'parent'       => $variation->parent->id];

            if (!wceb_is_bookable($variation)) {
                return $data;
            }

            $variations = $siblings;
            $variations = array_map(function ($v) {
                return "'" . esc_sql($v) . "'";
            },
                $variations
            );

            $variations = implode(',', $variations);
            $fill = implode(', ', array_fill(0, count($siblings), '%s'));

            $item_ids = $wpdb->get_results($wpdb->prepare("
        	SELECT order_item_id 
			FROM " . $wpdb->prefix . "woocommerce_order_itemmeta 
			WHERE meta_key='_variation_id' 
				AND meta_value IN ($fill) 
				AND order_item_id IN ( SELECT order_item_id 
										FROM " . $wpdb->prefix . "woocommerce_order_itemmeta 
										WHERE meta_key='_ebs_end_format' 
											AND meta_value>=%s )
			", array_merge($siblings, [date('Y-m-d')]))
            );

            foreach ($item_ids as $key => $item) {
                $data['booked'][] = [
                    'start_hour' => wc_get_order_item_meta($item->order_item_id, '_ebs_start_hour_display', true),
                    'end_hour'   => wc_get_order_item_meta($item->order_item_id, '_ebs_end_hour_display', true),
                    'start_date' => wc_get_order_item_meta($item->order_item_id, '_ebs_start_format', true),
                    'end_date'   => wc_get_order_item_meta($item->order_item_id, '_ebs_end_format', true),
                    'qty'        => wc_get_order_item_meta($item->order_item_id, '_qty', true),
                ];
            }

            return $data;
        }

        public function jebe_get_config_data()
        {
            global $post;

            $options = get_option('jebe_settings');
            $holidays = get_option('jebe_holidays');

            $messages = [
                'max_bookable_qty'                 => __('Max bookable quantity: {{qty}}', 'j_easy_booking_extended'),
                'date_range_error'                 => __('Please choose a valid range', 'j_easy_booking_extended'),
                'min_2_hours_error'                => __('The minimun bookable period is two hours', 'j_easy_booking_extended'),
                'range_overlaps_unavailable_hours' => __('There is no complete availability for selected range', 'j_easy_booking_extended'),
                'endtime_not_less_than_starttime'  => __('Check out date must be greater than check in ', 'j_easy_booking_extended'),
                'start_date_past'                  => __('Check in date must not be past ', 'j_easy_booking_extended'),
            ];

            sort($holidays);
            $today = date('Y-m-d');
            $holidays = array_filter($holidays, function ($value) {
                return $value >= date('Y-m-d');
            });

            // get hours not available for current variations

            $availabilities = [];

            // If not on a single product page, return
            if (is_product()) {

                $product = wc_get_product($post->ID);

                if ($product) {
                    // Load scripts only on product page if "booking" option is checked
                    if (wceb_is_bookable($product)) {

                        // WooCommerce 2.7 compatibility
                        if (is_callable([$product, 'get_type'])) {
                            $product_type = $product->get_type();
                        } else {
                            $product_type = $product->product_type;
                        }
                    }

                    // This should only work with variations, so this is ok
                    // other wise, we must take care of different product types
                    $variation_ids = $product->get_children();

                    if ($variation_ids) {

                        foreach ($variation_ids as $variation_id) {

                            $data = $this->get_variation_bookings($variation_id);

                            $availabilities[$variation_id] = $data;

                        }
                    }
                }
            }

            return [
                'allow_holiday_bookings' => $options['jebe_allow_book_in_holiday'],
                'holidays'               => $holidays,
                'availabilities'         => $availabilities,
                'options'                => $options,
                'messages'               => $messages];
        }

        public function jebe_get_booking_extras()
        {
            global $post;

            $extra_groups = get_option('jebe_extra_groups');

            $booking_extras = [];

            // If not on a single product page, return
            if (is_product()) {

                $product = wc_get_product($post->ID);

                if ($product) {

                    // This should only work with variations, so this is ok
                    // other wise, we must take care of different product types
                    $variation_ids = $product->get_children();

                    if ($variation_ids) {

                        foreach ($variation_ids as $variation_id) {

                            if (version_compare(WC_VERSION, '2.7', '<')) {
                                $product_extra_group_name = get_post_meta($variation_id, '_booking_extra_group', true);
                            } else {
                                $product_extra_group_name = $product->get_meta('_booking_extra_group', true);
                            }

                            foreach ($extra_groups as $key => $extra_group) {
                                if ($extra_group['extra_group_name'] === $product_extra_group_name) {
                                    $booking_extras[$variation_id] = $extra_group;
                                    break;
                                }
                            }

                        }
                    }
                }
            }

            if (is_null($booking_extras)) {
                return;
            }

            $booking_extras['product_id'] = $product->get_id();

            return $booking_extras;
        }

        public function jebe_get_new_price()
        {

            check_ajax_referer('set-dates', 'security');

            $options = get_option('jebe_settings');

            $min_bookable_hour = sprintf("%02d", $options['jebe_min_bookable_hour']) . ':00';
            $max_bookable_hour = sprintf("%02d", $options['jebe_max_bookable_hour']) . ':00';

            $product_id = isset($_POST['product_id']) ? absint($_POST['product_id']) : ''; // Product ID
            $variation_id = isset($_POST['variation_id']) ? absint($_POST['variation_id']) : ''; // Variation ID
            $children = isset($_POST['children']) ? array_map('absint', $_POST['children']) : []; // Product children for grouped and variable products

            $id = !empty($variation_id) ? $variation_id : $product_id; // Product or variation id

            $calc_mode = $this->options['easy_booking_calc_mode']; // Calculation mode (Days or Nights)

            $start_date = isset($_POST['start']) ? sanitize_text_field($_POST['start']) : ''; // Booking start date
            $end_date = isset($_POST['end']) ? sanitize_text_field($_POST['end']) : ''; // Booking end date

            $start = isset($_POST['start_format']) ? sanitize_text_field($_POST['start_format']) : ''; // Booking start date 'yyyy-mm-dd'
            $end = isset($_POST['end_format']) ? sanitize_text_field($_POST['end_format']) : ''; // Booking end date 'yyyy-mm-dd'

            $start_hour = isset($_POST['start_hour'])
                ? (($_POST['start_hour'] != '') ? sanitize_text_field($_POST['start_hour']) : $min_bookable_hour)
                : $min_bookable_hour; // Booking start date. Defaults to 8:00 if not present
            $end_hour = isset($_POST['end_hour'])
                ? (($_POST['end_hour'] != '') ? sanitize_text_field($_POST['end_hour']) : $max_bookable_hour)
                : $max_bookable_hour; // Booking end date. Defaults to 20:00 if not present

            $product = wc_get_product($product_id); // Product object
            $_product = ($product_id !== $id) ? wc_get_product($id) : $product; // Product or variation object


            // variation booking prices. This should be populated with defaults from admin.

            $default_prices = get_option('jebe_settings');
            $allow_book_in_holidays = $default_prices['jebe_allow_book_in_holiday'];

            $rates = get_option('jebe_rates');

            if (version_compare(WC_VERSION, '2.7', '<')) {
                $product_rate_name = get_post_meta($id, '_booking_rate', true);
            } else {
                $product_rate_name = $_product->get_meta('_booking_rate', true);
            }

            foreach ($rates as $key => $rate) {
                if ($rate['name'] === $product_rate_name) {
                    $product_rate = $rate;
                    break;
                }
            }

            $booking_prices['hour'] = $_prices['hour'] > 0 ? $_prices['hour'] : $booking_prices['hour'];
            $booking_prices['day'] = $_prices['day'] > 0 ? $_prices['day'] : $booking_prices['day'];
            $booking_prices['weekend'] = $_prices['weekend'] > 0 ? $_prices['weekend'] : $booking_prices['weekend'];
            $booking_prices['holiday'] = $_prices['holiday'] > 0 ? $_prices['holiday'] : $booking_prices['holiday'];

            if (is_null($product_rate)) {
                return;
            }

            if (!$_product) {
                return;
            }

            // If product is variable and no variation was selected
            if ($product->is_type('variable') && empty($variation_id)) {
                $this->wceb_throw_error(3);
            }

            // If product is grouped and no quantity was selected for grouped products
            if ($product->is_type('grouped') && empty($children)) {
                $this->wceb_throw_error(4);
            }

            $number_of_dates = wceb_get_product_booking_dates($_product);

            // If date format is "one", check only one date is set
            if ($number_of_dates === 'one') {

                $dates = 'one_date';
                $duration = 1;

                // If end date is set
                if (!empty($end_date) || !empty($end)) {
                    $this->wceb_throw_error(5);
                }

                // If date is empty
                if (empty($start_date) || empty($start)) {
                    $this->wceb_throw_error(6);
                }

            } else { // "Two" dates check

                $dates = 'two_dates';

                // If one date is empty
                if (empty($start_date) || empty($end_date) || empty($start) || empty($end)) {
                    $this->wceb_throw_error(2);
                }

                $start_time = strtotime($start);
                $end_time = strtotime($end);

                // If end date is before start date
                if ($end_time < $start_time) {
                    $this->wceb_throw_error(1);
                }

                // Get booking duration in days
                $duration = absint(($start_time - $end_time) / 86400);

                if ($duration == 0) {
                    $duration = 1;
                }

                // If booking mode is days and calculation mode is set to "Days", add one day
                if ($calc_mode === 'days' && ($start != $end)) {
                    $duration += 1;
                }

                $booking_duration = wceb_get_product_booking_duration($_product);

                // If booking mode is weeks and duration is a multiple of 7
                if ($booking_duration === 'weeks') {

                    if ($calc_mode === 'nights' && $duration % 7 === 0) { // If in weeks mode, check that the duration is a multiple of 7
                        $duration /= 7;
                    } else if ($calc_mode === 'days' && $duration % 6 === 0) { // Or 6 in "Days" mode
                        $duration /= 6;
                    } else { // Otherwise throw an error
                        $this->wceb_throw_error(1);
                    }

                } else if ($booking_duration === 'custom') {

                    $custom_booking_duration = wceb_get_product_custom_booking_duration($_product);

                    if ($duration % $custom_booking_duration === 0) {
                        $duration /= $custom_booking_duration;
                    } else {
                        $this->wceb_throw_error(1);
                    }

                }

                // If number of days is inferior to 0
                if ($duration <= 0) {
                    $this->wceb_throw_error(1);
                }

            }

            // Get additional costs (for WooCommerce Product Addons)
            $additional_cost = $this->jebe_get_additional_costs($_product);

            // Store data in array
            $data = [
                'start_date' => $start_date,
                'start'      => $start,
            ];

            if (isset($duration) && !empty($duration)) {
                $data['duration'] = $duration;
            }

            if (isset($end_date) && !empty($end_date)) {
                $data['end_date'] = $end_date;
            }

            if (isset($end) && !empty($end)) {
                $data['end'] = $end;
            }

            $booking_data = [];

            $new_price = 0;
            $new_regular_price = 0;

            // Grouped or Bundle product types
            if ($product->is_type('grouped') || $product->is_type('bundle')) {

                if (!empty($children))
                    foreach ($children as $child_id => $quantity) {

                        if ($quantity <= 0 || ($child_id === $id)) {
                            continue;
                        }

                        $child = wc_get_product($child_id);

                        $children_prices[$child_id] = (array)wceb_get_product_price($product, $child, false, 'array');

                        // Multiply price by duration only if children is bookable
                        if ($children_prices[$child_id]) {

                            if (wceb_is_bookable($child)) {

                                if ($children_prices[$child_id])
                                    foreach ($children_prices[$child_id] as $price_type => $price) {

                                        if ($price === "") {
                                            continue;
                                        }

                                        if ($number_of_dates === 'two') {
                                            $price *= $duration;
                                        }

                                        ${'child_new_' . $price_type} = apply_filters(
                                            'easy_booking_' . $dates . '_price',
                                            wc_format_decimal($price), // Regular or sale price for x days
                                            $product, $child, $data, $price_type
                                        );

                                    }

                            } else {

                                $child_new_price = wc_format_decimal($children_price[$child_id]['price']);

                                if (isset($children_price[$child_id]['regular_price'])) {
                                    $child_new_regular_price = wc_format_decimal($children_price[$child_id]['regular_price']);
                                }

                            }

                        } else {

                            // Tweak for not individually sold bundled products
                            $child_new_price = 0;
                            $child_new_regular_price = 0;

                        }

                        // Maybe add additional costs
                        if (isset($additional_cost[$child_id])) {

                            $child_new_price = $this->wceb_add_additional_costs($child_new_price, $additional_cost[$child_id], $duration);

                            if (isset($child_new_regular_price)) {
                                $child_new_regular_price = $this->wceb_add_additional_costs($child_new_regular_price, $additional_cost[$child_id], $duration);
                            }

                        }

                        $data['new_price'] = $child_new_price;

                        if (isset($child_new_regular_price) && !empty($child_new_regular_price)) {
                            $data['new_regular_price'] = $child_new_regular_price;
                        }

                        // Store parent produt for bundled items
                        if ($product->is_type('bundle')) {
                            $data['grouped_by'] = $product;
                        }

                        $booking_data[$child_id] = $data;

                        if ($product->is_type('grouped')) {
                            $new_price += wc_format_decimal($child_new_price * $quantity);

                            if (isset($child_new_regular_price)) {
                                $new_regular_price += wc_format_decimal($child_new_regular_price * $quantity);
                            }
                        }

                    }

                if ($product->is_type('bundle')) {

                    $prices = (array)wceb_get_product_price($product, false, false, 'array');

                    if ($prices)
                        foreach ($prices as $price_type => $price) {

                            if ($price === "") {
                                continue;
                            }

                            if ($number_of_dates === 'two') {
                                $price *= $duration;
                            }

                            ${'new_' . $price_type} = apply_filters(
                                'easy_booking_' . $dates . '_price',
                                wc_format_decimal($price), // Regular or sale price for x days
                                $product, $_product, $data, $price_type
                            );

                            if (isset($additional_cost[$id]) && $additional_cost[$id] > 0) {
                                ${'new_' . $price_type} = $this->wceb_add_additional_costs(${'new_' . $price_type}, $additional_cost[$id], $duration);
                            }

                        }

                }

            } else {

                // Get product price and (if on sale) regular price
                $prices = (array)wceb_get_product_price($_product, false, false, 'array');

                if ($prices)
                    foreach ($prices as $price_type => $price) {

                        if ($price === "") {
                            continue;
                        }

                        if ($number_of_dates === 'two') {
                            $price *= $duration;
                        }

                        ${'new_' . $price_type} = apply_filters(
                            'easy_booking_' . $dates . '_price',
                            wc_format_decimal($price), // Regular or sale price for x days
                            $product, $_product, $data, $price_type
                        );

                        if (isset($additional_cost[$id]) && $additional_cost[$id] > 0) {
                            ${'new_' . $price_type} = $this->jebe_add_additional_costs(${'new_' . $price_type}, $additional_cost[$id], $duration);
                        }

                    }

            }


            // Hour based price calculation
            if (wceb_is_bookable($product)) {

                $start_time = DateTime::createFromFormat('H:i Y-m-d', $start_hour . ' ' . $start);
                $end_time = DateTime::createFromFormat('H:i Y-m-d', $end_hour . ' ' . $end);

                // Same day. Hour rate
                if ($duration == 1 && ($start == $end)) {

                    if (!empty($start_hour) && !empty($end_hour)) {

                        $diff_time = $end_time->diff($start_time);
                        $booking_duration = ($diff_time->days * 24) + $diff_time->h + ($diff_time->i / 60) + ($diff_time->s / 3600);

                        // apply day rate if booking is 8 or more hours ( maybe moved to settings )
                        if ($booking_duration >= 8) {
                            $new_price = $product_rate['rates']['hour_8'] + (($booking_duration - 8) * $product_rate['rates']['hourextra']);
                        } else if ($booking_duration >= 4) {
                            $new_price = $product_rate['rates']['hour_4'] + (($booking_duration - 4) * $product_rate['rates']['hourextra']);
                        } else if ($booking_duration >= 2) {
                            $new_price = $product_rate['rates']['hour_2'] + (($booking_duration - 2) * $product_rate['rates']['hourextra']);
                        } else {
                            // less than 2 hours not allowed
                            return;
                        }

                        $data['boking_duration'] = $booking_duration;
                        $data['duration'] = $duration;

                    }
                } else {

                    if (jebe_booking_is_weekend($start_time, $end_time)) {

                        $new_price = $booking_prices['weekend'];

                    } else {

                        // price with full day range
                        $new_price = $this->jebe_get_price_for_days($duration, $product_rate['rates']);
                        //invalid range. Should never happen
                        if ($new_price === false) {
                            return;
                        }

                        $data['price_without_extra_hours'] = $new_price;

                        //price with days-2 + extra hours

                        //hours until day 2
                        $next_opening_date = DateTime::createFromFormat('H:i Y-m-d', $min_bookable_hour . ' ' . $start)->add(new DateInterval('P1D'));
                        $booking_start = DateTime::createFromFormat('H:i Y-m-d', $start_hour . ' ' . $start);
                        $diff_time = $next_opening_date->diff($booking_start);
                        $first_day_duration = ($diff_time->days * 24) + $diff_time->h + ($diff_time->i / 60) + ($diff_time->s / 3600);
                        $first_day_price = $first_day_duration * $product_rate['rates']['hourextra'];

                        //hours until last day
                        if ($duration == 1) {
                            // if booking spreads in two days, only need to count hours from second day opening because previous nightly hours have been counted previously
                            $min_bookable_date = DateTime::createFromFormat('H:i Y-m-d', $min_bookable_hour . ' ' . $end);
                        } else {
                            $min_bookable_date = DateTime::createFromFormat('H:i Y-m-d', $max_bookable_hour . ' ' . $end)->sub(new DateInterval('P1D'));
                        }
                        $booking_end = DateTime::createFromFormat('H:i Y-m-d', $end_hour . ' ' . $end);
                        $diff_time = $booking_end->diff($min_bookable_date);
                        $last_day_duration = ($diff_time->days * 24) + $diff_time->h + ($diff_time->i / 60) + ($diff_time->s / 3600);
                        $last_day_price = $last_day_duration * $product_rate['rates']['hourextra'];

                        //price for duration-2
                        $price = $this->jebe_get_price_for_days($duration - 2, $product_rate['rates']);

                        $price_with_extra_hours = $price + $first_day_price + $last_day_price;

                        $data['price_with_extra_hours'] = $price_with_extra_hours;

                        if ($price_with_extra_hours < $new_price) {
                            $new_price = $price_with_extra_hours;
                        }

                    }

                }
            }

            if (isset($_POST['additional_cost'])) {
                $new_price += $_POST['additional_cost'][$_POST['variation_id']];
                $data['extras'] = $_POST['extras'];
            }

            $data['new_price'] = $new_price;

            if (isset($new_regular_price) && !empty($new_regular_price) && ($new_regular_price !== $new_price)) {
                $data['new_regular_price'] = $new_regular_price;
            } else {
                unset($data['new_regular_price']); // Unset value in case it was set for a child product
            }
            $data['start_hour'] = $start_hour;
            $data['end_hour'] = $end_hour;

            $data['booking_prices'] = $booking_prices;

            $data['_star'] = $_POST['start_hour'];

            $data['fragments'] = apply_filters('easy_booking_fragments', [
                    'session'               => true,
                    'booking_price'         => esc_attr($new_price),
                    'booking_regular_price' => '',
                    'input.wceb_nonce'      => '<input type="hidden" name="_wceb_nonce" class="wceb_nonce" value="' . wp_create_nonce('set-dates') . '">',
                ]
            );

            if (isset($details)) {
                $data['fragments']['p.booking_details'] = wp_kses_post($details);
            }

            $data['max_bookable_qty'] = $this->get_max_bookable_qty($variation_id, $data);

            $booking_data[$id] = $data;


            // Update session data
            if (!WC()->session->has_session()) {
                WC()->session->set_customer_session_cookie(true);
            }

            WC()->session->set('booking', $booking_data);
            WC()->session->set('booking_hours', $booking_data);

            wp_send_json($data);
            die();
        }

        private function jebe_get_price_for_days($duration, $rate)
        {
            $price = 0;

            if ($duration >= 82) {
                $price = $duration * $rate['days_82'];
            } else if ($duration >= 65) {
                $price = $duration * $rate['days_65_81'];
            } else if ($duration >= 50) {
                $price = $duration * $rate['days_50_64'];
            } else if ($duration >= 37) {
                $price = $duration * $rate['days_37_49'];
            } else if ($duration >= 26) {
                $price = $duration * $rate['days_26_36'];
            } else if ($duration >= 17) {
                $price = $duration * $rate['days_17_25'];
            } else if ($duration >= 10) {
                $price = $duration * $rate['days_10_16'];
            } else if ($duration >= 5) {
                $price = $duration * $rate['days_5_9'];
            } else if ($duration >= 3) {
                $price = $duration * $rate['days_3_4'];
            } else if ($duration === 2) {
                $price = $rate['day_2'];
            } else if ($duration === 1) {
                $price = $rate['day_1'];
            } else {
                $price = false;
            }

            return $price;
        }

        public function holiday_count($start_date, $end_date = null)
        {

            $holidays = get_option('jebe_holidays');

            if (is_null($end_date)) {

                if (in_array($start_date->format('Y-m-d'), $holidays))
                    return 1;
                else
                    return 0;

            }


            $interval = new DateInterval('P1D');
            $date_range = new DatePeriod($start_date, $interval, $end_date);

            $count = 0;

            foreach ($date_range as $date) {

                if (in_array($date->format('Y-m-d'), $holidays)) {
                    $count++;
                }
            }

            return $count;

        }

        /**
         *
         * Adds additional costs (compatibility with WooCommerce Product Addons)
         *
         * @param int - $price - Product price
         * @param int - $additional_cost
         * @param int - $duration
         *
         * @return int - $price
         *
         **/
        private function jebe_add_additional_costs($price, $additional_cost, $duration)
        {

            // Pass true to filter to multiply additional costs by booking duration (default: false)
            if (true === apply_filters('easy_booking_multiply_additional_costs', false)) {
                $price += ($additional_cost * $duration);
            } else {
                $price += $additional_cost;
            }

            return $price;
        }

        /**
         *
         * Formats additional costs (compatibility with WooCommerce Product Addons)
         *
         * @param Wc_Product - $_product - Product or variation object
         *
         * @return array - $additional_cost
         *
         **/
        private function jebe_get_additional_costs($_product)
        {

            // Get additional costs
            $additional_cost = isset($_POST['additional_cost']) ? array_map('wc_format_decimal', $_POST['additional_cost']) : [];

            $prices_include_tax = get_option('woocommerce_prices_include_tax');
            $tax_display_mode = get_option('woocommerce_tax_display_shop');

            // Get additional costs including or excluding taxes (for WooCommerce Product Addons)
            if (!empty($additional_cost)) {

                foreach ($additional_cost as $ac_id => $ac_amount) {

                    if ($_product->is_taxable()) {

                        $rates = WC_Tax::get_base_tax_rates($_product->get_tax_class());

                        if ($prices_include_tax === 'yes' && $tax_display_mode === 'excl') {

                            $taxes = WC_Tax::calc_exclusive_tax($ac_amount, $rates);

                            if ($taxes)
                                foreach ($taxes as $tax) {
                                    $additional_cost[$ac_id] += $tax;
                                }

                        } else if ($prices_include_tax === 'no' && $tax_display_mode === 'incl') {

                            $taxes = WC_Tax::calc_inclusive_tax($ac_amount, $rates);

                            if ($taxes)
                                foreach ($taxes as $tax) {
                                    $additional_cost[$ac_id] -= $tax;
                                }

                        }

                    }

                }

            }

            return $additional_cost;
        }

        public function remove_ajax_actions()
        {
            global $wp_filter;
            $tag = 'wp_ajax_add_new_price';
            $priority = 10;
            $tag_hooks = $wp_filter[$tag]->callbacks[$priority];
            if (is_null($tag_hooks))
                return;
            foreach ($tag_hooks as $the_tag => $the_callback) {
                if ($the_callback['function'][1] == 'wceb_get_new_price') {
                    remove_action($tag, $the_callback['function'], $priority);
                }
            }
        }

        public function remove_actions()
        {
            global $wp_filter;
            $tag = 'woocommerce_single_variation';
            $priority = 10;
            $tag_hooks = $wp_filter[$tag]->callbacks[$priority];
            if (is_null($tag_hooks))
                return;
            foreach ($tag_hooks as $the_tag => $the_callback) {
                if ($the_callback['function'][1] == 'wceb_before_add_to_cart_button') {
                    remove_action($tag, $the_callback['function'], $priority);
                }
            }
            $tag = 'woocommerce_before_add_to_cart_button';
            $priority = 20;
            $tag_hooks = $wp_filter[$tag]->callbacks[$priority];
            if (is_null($tag_hooks))
                return;
            foreach ($tag_hooks as $the_tag => $the_callback) {
                if ($the_callback['function'][1] == 'wceb_before_add_to_cart_button') {
                    remove_action($tag, $the_callback['function'], $priority);
                }
            }
        }

        /**
         *
         * Check if WooCommerce Easy Booking is active
         *
         * @return bool
         *
         **/
        private function jebe_easy_booking_is_active()
        {

            // Get active plugins
            $active_plugins = (array)get_option('active_plugins', []);

            if (is_multisite()) {
                $active_plugins = array_merge($active_plugins, get_site_option('active_sitewide_plugins', []));
            }

            // Return true if Easy Booking is active
            return (array_key_exists('j-easy-booking-extended/j-easy-booking-extended.php', $active_plugins) || in_array('j-easy-booking-extended/j-easy-booking-extended.php', $active_plugins));

        }

        /**
         *
         * Init plugin
         *
         **/
        public function jebe_init()
        {

            // Don't init plugin if WooCommerce Easy Booking isn't at least version 1.9.
            if (!method_exists(WCEB(), 'wceb_get_version')) {
                return;
            }

            // Define constants
            $this->jebe_define_constants();

            // load plugin textdomain
            $this->jebe_load_textdomain();

            // Common includes
            $this->jebe_includes();

            // Admin includes
            if (is_admin()) {
                $this->jebe_admin_includes();
            }

            // Frontend includes
            if (!is_admin() || defined('DOING_AJAX')) {
                $this->jebe_frontend_includes();
            }

        }

        /**
         *
         * Define constants
         *
         **/
        private function jebe_define_constants()
        {
            // Plugin directory path
            define('JEBE_PLUGIN_FILE', __FILE__);
        }

        /**
         *
         * Load plugin textdomain for translations
         *
         **/
        private function jebe_load_textdomain()
        {
            load_plugin_textdomain('j_easy_booking_extended', false, basename(dirname(__FILE__)) . '/languages/');
        }

        /**
         *
         * Common includes
         *
         **/
        private function jebe_includes()
        {
            include_once('core/jebe-functions.php');
            include_once('core/front/cart.php');
            /*include_once( 'includes/jebe-functions.php' );
            include_once( 'includes/jebe-date-functions.php' );
            include_once( 'includes/jebe-product-functions.php' );
            include_once( 'includes/class-jebe-ajax.php' );
            include_once( 'includes/class-jebe-cart.php' );
            include_once( 'includes/admin/class-jebe-order.php' );*/
        }

        /**
         *
         * Admin includes
         *
         **/
        private function jebe_admin_includes()
        {
            include_once('core/admin.php');
            /*
            include_once( 'includes/admin/class-jebe-admin-assets.php' );
            include_once( 'includes/settings/class-jebe-settings.php' );
            include_once( 'includes/imports/class-jebe-list-imports.php' );
            include_once( 'includes/imports/class-jebe-imports.php' );
            include_once( 'includes/reports/class-jebe-reports.php' );
            */
        }

        /**
         *
         * Frontend includes
         *
         **/
        private function jebe_frontend_includes()
        {
            include_once('core/front/functions.php');
            include_once('core/front/cart.php');

            //include_once( 'includes/class-jebe-product-view.php' );
            //include_once( 'includes/class-jebe-assets.php' );
        }

        /**
         *
         * Add notices
         *
         **/
        public function jebe_add_notices()
        {

            if (is_multisite()) {
                return;
            }

            if (!method_exists(WCEB(), 'wceb_get_version')) {
                include_once('includes/admin/views/html-jebe-notice-update.php');

                return;
            }

            $settings = get_option('easy_booking_availability_settings');
            $license_key = isset($settings['easy_booking_availability_license_key']) ? $settings['easy_booking_availability_license_key'] : false;

            if (empty($license_key)) {
                $license_key = false;
            }

            $screen = get_current_screen();

            // Don't display notices on the plugin settings page
            if (in_array($screen->id, ['easy-booking_page_easy-booking-availability-check'])) {
                return;
            }

            // Display a notice to init availabilities
            if (get_option('easy_booking_display_notice_init') !== '1') {
                include_once('includes/admin/views/html-jebe-notice-init.php');
            }

            if (get_option('easy_booking_display_notice_init') == '1' && get_option('easy_booking_display_notice_init_after_update') !== '1') {
                include_once('includes/admin/views/html-jebe-notice-init-after-update.php');
            }

            // Display license key notice
            if (!is_multisite() && get_option('easy_booking_display_notice_jebe_license') !== '1' && !$license_key) {
                include_once('includes/admin/views/html-jebe-notice-license-key.php');
            }
        }


        // WCEB Private functions

        /**
         *
         * Throws an error message
         *
         * @param int $error_code
         *
         **/
        private function wceb_throw_error($error_code)
        {
            $error_message = $this->wceb_get_date_error($error_code);
            wc_add_notice($error_message, 'error');

            $this->wceb_error_fragment($error_message);
            $this->wceb_clear_booking_session();
            die();
        }

        /**
         *
         * Gets error messages
         *
         * @param int $error_code
         *
         * @return str $err - Error message
         *
         **/
        private function wceb_get_date_error($error_code)
        {

            switch ($error_code) {
                case 1:
                    $err = __('Please choose valid dates', 'easy_booking');
                    break;
                case 2:
                    $err = __('Please choose two dates', 'easy_booking');
                    break;
                case 3:
                    $err = __('Please select product option', 'easy_booking');
                    break;
                case 4:
                    $err = __('Please choose the quantity of items you wish to add to your cart&hellip;', 'woocommerce');
                    break;
                case 5:
                    $err = __('You can only select one date', 'easy_booking');
                    break;
                case 6:
                    $err = __('Please choose a date', 'easy_booking');
                    break;
                default:
                    $err = '';
                    break;
            }

            return $err;
        }

        /**
         *
         * Updates error messages with Ajax
         *
         * @param str $error_message
         *
         **/
        public function wceb_error_fragment($error_message)
        {

            header('Content-Type: application/json; charset=utf-8');

            ob_start();
            wc_print_notices();
            $error_message = ob_get_clean();

            $data = [
                'errors' => [
                    'div.wc_ebs_errors' => '<div class="wc_ebs_errors">' . $error_message . '</div>',
                ],
            ];

            wp_send_json($data);

            die();

        }

        /**
         *
         * Clears session if "Clear" button is clicked on the calendar
         *
         **/
        public function wceb_clear_booking_session()
        {

            check_ajax_referer('set-dates', 'security');

            $session = WC()->session->get('booking');

            if (!empty($session)) {
                WC()->session->set('booking', '');
            }

            $session_set = false;

            die($session_set);
        }


    }

    function JEBE()
    {
        return J_Easy_Booking_Extended::instance();
    }

    JEBE();

endif;
